/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "gtkeditor.h"
#include "internal.h"

/* widget stuff */
static void   gtk_editor_class_init          (GtkEditorClass   *klass);
static void   gtk_editor_init                (GtkEditor        *editor);
static void   gtk_editor_destroy             (GtkObject        *object);

/* hilit */


/* indent */
static void   default_one_line_ind           (GtkEditor        *editor);

/* events */
static gint   gtk_editor_key_press           (GtkWidget        *widget,
					      GdkEventKey      *event);

/* --<local data>--------------------------------------------------------- */
static GtkWidgetClass *parent_class = NULL;

/* --<widget initialization, constructor and destructor>------------------ */
guint
gtk_editor_get_type ()
{
  static guint editor_type = 0;

  if (!editor_type)
    {
      GtkTypeInfo editor_info =
      {
	"GtkEditor",
	sizeof (GtkEditor),
	sizeof (GtkEditorClass),
	(GtkClassInitFunc) gtk_editor_class_init,
	(GtkObjectInitFunc) gtk_editor_init,
	(GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL,
      };

      editor_type = gtk_type_unique (gtk_text_get_type (), &editor_info);
    }

  return editor_type;
}

static void
gtk_editor_class_init (GtkEditorClass *class)
{
  GtkObjectClass *object_class;
  GtkEditableClass *editable_class;

  object_class = (GtkObjectClass*) class;
  editable_class = (GtkEditableClass *) class;

  /* init local data */
  parent_class = gtk_type_class (gtk_text_get_type ());

  /* setup signals */
  object_class->destroy = gtk_editor_destroy;

  editable_class->insert_text = gtk_editor_insert_text;
  editable_class->delete_text = gtk_editor_delete_text;
}

static void
gtk_editor_init (GtkEditor *editor)
{
  /* hilite */
  editor->n_text_properties = 0;
  editor->text_properties = NULL;
  editor->n_token_entries = 0;
  editor->token_entries = NULL;
  editor->n_block_entries = 0;
  editor->block_entries = NULL;
  editor->tokens = NULL;
  editor->cur_token = NULL;
  editor->cur_token_pos = 0;
  editor->blocks = NULL;
  editor->cur_block = NULL;
  editor->prev_block = NULL;
  editor->next_block = NULL;

  gtk_signal_connect (GTK_OBJECT (editor), "key_press_event",
		      GTK_SIGNAL_FUNC (gtk_editor_key_press),
		      NULL);

  /* indent */
  editor->one_line_ind = (GtkEditorIndFunc)default_one_line_ind;
  editor->ind_accel_key = GDK_Tab;
  editor->ind_accel_mod = 0;
  editor->indenter_data = NULL;
}

GtkWidget*
gtk_editor_new (GtkAdjustment *hadj,
		GtkAdjustment *vadj)
{
  GtkWidget *editor;

  if (hadj)
    g_return_val_if_fail (GTK_IS_ADJUSTMENT (hadj), NULL);
  if (vadj)
    g_return_val_if_fail (GTK_IS_ADJUSTMENT (vadj), NULL);
  
  editor = gtk_widget_new (GTK_TYPE_EDITOR,
			   "hadjustment", hadj,
			   "vadjustment", vadj,
			   NULL);

  return GTK_WIDGET (editor);
}

static void
gtk_editor_destroy (GtkObject *object)
{
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_EDITOR (object));

  GTK_OBJECT_CLASS(parent_class)->destroy (object);
}

/* --<hilite stuff>------------------------------------------------------- */
GdkFont*
gtk_editor_fontdup (GdkFont *font)
{
  if (font) {
    gdk_font_ref (font);
  }
  return font;
}

GdkColor*
gtk_editor_coldup (const GdkColor *color)
{
  GdkColor *new;
  if (color) {
    new = g_new (GdkColor, 1);
    memcpy (new, color, sizeof (GdkColor));
  } else {
    new = NULL;
  }
  return new;
}

/* --<indent stuff>------------------------------------------------------- */
/* FIXME: we should have some better api */
static void
default_one_line_ind (GtkEditor *editor)
{
  guint pos = gtk_editable_get_position (GTK_EDITABLE (editor));

  gtk_text_freeze (GTK_TEXT (editor));
  /* FIXME: change this when they change the interface to *_move_* */
  gtk_signal_emit_by_name (GTK_OBJECT (editor), "move_to_column", 0);
  gtk_text_set_point (GTK_TEXT (editor),
		      gtk_editable_get_position (GTK_EDITABLE (editor)));
  gtk_text_insert (GTK_TEXT (editor), NULL, NULL, NULL, "\t", 1);
  gtk_text_thaw (GTK_TEXT (editor));
  gtk_editable_set_position (GTK_EDITABLE (editor), pos + 1); /* old pos + tab */
}

void
gtk_editor_install_indenter (GtkEditor *editor, GtkEditorIndFunc ind,
			     guint accel_key, GdkModifierType accel_mod,
			     gpointer user_data)
{
  editor->one_line_ind = ind;
  editor->ind_accel_key = accel_key;
  editor->ind_accel_mod = accel_mod;
  editor->indenter_data = user_data;
}

/* --<event handles>------------------------------------------------------ */
static gint
gtk_editor_key_press (GtkWidget *widget, GdkEventKey *event)
{
  GtkEditor *editor;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (event != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_EDITOR (widget), FALSE);

  editor = GTK_EDITOR (widget);

  if (editor->one_line_ind) {
    if ( (!editor->ind_accel_mod || (event->state & editor->ind_accel_mod))
	 && (event->keyval == editor->ind_accel_key) ) {

      gtk_signal_emit_stop_by_name (GTK_OBJECT (widget), "key_press_event");
      editor->one_line_ind (editor, editor->indenter_data);
      return TRUE;
    }
  }

  return FALSE;
}
